"""
    QuaO Project ibm_cloud_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_ibm_runtime import QiskitRuntimeService

from ...enum.provider_type import ProviderType
from ...model.provider.provider import Provider
from ...config.logging_config import logging

class IbmCloudProvider(Provider):

    def __init__(self, api_key, crn):
        super().__init__(ProviderType.IBM_CLOUD)
        self.api_key = api_key
        self.crn = crn
        self.channel = 'ibm_cloud'

    def get_backend(self, device_specification: str):
        """

        @param device_specification:
        @return:
        """
        provider = self.collect_providers()

        return provider.get_backend(name=device_specification)

    def collect_providers(self):
        """

        @return:
        """

        logging.info('Connect to Ibm Cloud provider')
        return QiskitRuntimeService(channel=self.channel, token=self.api_key, instance=self.crn)



"""
    QuaO Project quao_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_aer import Aer

from ...enum.provider_type import ProviderType
from ...model.provider.provider import Provider
from braket.devices import LocalSimulator
from ...config.logging_config import logging

class QuaoProvider(Provider):
    def __init__(self):
        super().__init__(ProviderType.QUAO_QUANTUM_SIMULATOR)

    def get_backend(self, device_specification):
        """

        @param device_specification:
        @return:
        """

        providers = self.collect_providers()
    
        aer_device_names = set(map(self.__map_aer_backend_name, providers[0].backends()))
        aws_device_names = providers[1].registered_backends()

        if aer_device_names.__contains__(device_specification):
            return providers[0].get_backend(device_specification)

        if aws_device_names.__contains__(device_specification):
            return LocalSimulator(device_specification)

        raise Exception('Unsupported device')

    def collect_providers(self):
        """

        @return:
        """
        logging.info('Connect to Quao provider')
        return [Aer, LocalSimulator()]

    @staticmethod
    def __map_aer_backend_name(backend):
        return backend.configuration().backend_name



"""
    QuaO Project job_status.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit import QiskitError
from qiskit_ibm_runtime import QiskitRuntimeService
from qiskit.providers.jobstatus import JobStatus
from json import dumps

from ...data.job.job_response import JobResponse
from ...enum.media_type import MediaType
from ...util.json_parser_util import JsonParserUtils


class Job:
    def __init__(self, channel, token, crn, provider_job_id):
        self.channel = channel if channel else "ibm_quantum"
        self.token = token if token else ""
        self.crn = crn if crn else None
        self.job_id = provider_job_id if provider_job_id else ""
        self.retrieve_job = self.__get_job()

    def fetch_job(self) -> JobResponse:

        content_type = None
        job_histogram = None
        job_result = None
        status = self.retrieve_job.status()
        job_status = status.name

        if status == JobStatus.DONE:
            try:
                result = self.retrieve_job.result()
                # case function `sampler`
                if self.retrieve_job.program_id == 'sampler':
                    job_result = JsonParserUtils.parse(result.__dict__)
                # case function `circuit-runner`

                else:
                    job_result = JsonParserUtils.parse(result.to_dict())
                    # handle case Statevector object not serializable when run with backend `statevector_simulator`
                    if self.retrieve_job.backend().backend_name == 'statevector_simulator':
                        job_result = dumps(job_result, default=str)
                content_type = MediaType.APPLICATION_JSON.value

                # handle circuit has measure
                try:
                    job_histogram = result.get_counts()
                except QiskitError:
                    job_histogram = None

            except Exception as exception:
                job_result = {
                    "error": "Exception when get job result",
                    "exception": str(exception)
                }
                job_status = JobStatus.ERROR.value

        return JobResponse(
            provider_job_id=self.job_id,
            job_status=job_status,
            job_result=job_result,
            content_type=content_type,
            job_histogram=job_histogram
        )

    def __get_job(self):
        service = QiskitRuntimeService(channel=self.channel,
                                       token=self.token,
                                       instance=self.crn)

        return service.job(job_id=self.job_id)



"""
    QuaO Project quao_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit import transpile
from qiskit import QiskitError

from ...enum.job_status import JobStatus
from ...util.json_parser_util import JsonParserUtils
from ...enum.sdk import Sdk
from ..device.device import Device
from ..provider.provider import Provider
from ...config.logging_config import logging

class QuaoDevice(Device):
    def __init__(self, provider: Provider, device_specification: str, sdk: str):
        super().__init__(provider, device_specification)
        self.sdk = sdk

    def _create_job(self, circuit, shots):
        logging.info('Create Quao job with {0} shots'.format(shots))
        if Sdk.QISKIT.value.__eq__(self.sdk):
            transpiled_circuit = transpile(circuits=circuit, backend=self.device)

            return self.device.run(transpiled_circuit, shots=shots)

        if Sdk.BRAKET.value.__eq__(self.sdk):
            return self.device.run(task_specification=circuit, shots=shots)

        raise Exception("Sdk not supported!")

    def _is_simulator(self) -> bool:
        return True

    def _parse_job_result(self, job_result) -> dict:
        if Sdk.QISKIT.value.__eq__(self.sdk):
            return JsonParserUtils.parse(job_result.to_dict())

        if Sdk.BRAKET.value.__eq__(self.sdk):
            return JsonParserUtils.parse(job_result.__dict__)

    def _produce_histogram_data(self, job_result) -> dict:
        if Sdk.QISKIT.value.__eq__(self.sdk):
            try:
                histogram_data = job_result.get_counts()
            except QiskitError as qiskit_error:
                logging.info("Can't produce histogram with error: {0}".format(str(qiskit_error)))
                histogram_data = None

            return histogram_data

        if Sdk.BRAKET.value.__eq__(self.sdk):
            return dict(job_result.measurement_counts)

    def _get_provider_job_id(self, job) -> str:
        if Sdk.QISKIT.value.__eq__(self.sdk):
            return job.job_id()

        if Sdk.BRAKET.value.__eq__(self.sdk):
            return job.id

    def _get_job_status(self, job) -> str:
        if Sdk.QISKIT.value.__eq__(self.sdk):
            return job.status().name

        if Sdk.BRAKET.value.__eq__(self.sdk):
            job_state = job.state()
            if JobStatus.COMPLETED.value.__eq__(job_state):
                job_state = JobStatus.DONE.value
            return job_state
